/*
 * SWCPU.cpp
 *
 *  Created on: Jun 15, 2011
 *      Author: fritz
 */

#include "SWCPU.h"

// Logging:
ILog * _log = 0;

#ifdef _WIN32
#define dllexport  __declspec(dllexport)
#else
#define dllexport
#endif

extern "C" dllexport int Cookie() {
	return 0x10201130;
}

extern "C" dllexport void SetLog(ILog * log) {
	_log = log;
}

IConfig * _config = 0;
extern "C" dllexport void SetConfig(IConfig * config) {
	_config = config;
}

extern "C" dllexport IAlignment * CreateAlignment(int const gpu_id) {
	return new SWCPUCor(gpu_id);
}

extern "C" dllexport void ExternalDeleteString(char* mem) {
	delete[] mem;
}

extern "C" dllexport void DeleteAlignment(SWCPUCor* instance) {
	delete instance;
}

void SWCPUCor::Allocate() {
	local_mat_line = new short[corr_length];
}

void SWCPUCor::FreeMem() {
	if (!type) {
		delete[] mat_pointer;
		delete[] result;
	}
	delete[] local_mat_line;
}

SWCPUCor::SWCPUCor(int gpu_id) {
	sse = new SSEAligner();

	read_length = Config.GetInt("qry_max_len"); //Read Size
	corr_length = Config.GetInt("corridor")+1;//Corridor Size
	ref_length = read_length + corr_length; //Reference Size

	cigar = bool (((gpu_id >> 8) & 0xFF )==1);

	batch_size=1;

	mat = Config.GetFloat("score_match");
	mis = Config.GetFloat("score_mismatch");
	gap_read = Config.GetFloat("score_gap_read");
	gap_ref = Config.GetFloat("score_gap_ref");

	type = true;
	short temp[6][6]= {
			mat,mis,mis,mis,0,mis,
			mis,mat,mis,mis,0,mis,
			mis,mis,mat,mis,0,mis,
			mis,mis,mis,mat,0,mis,
			0,0,0,0,0,0,
			0  ,  0,  0,  0,  0,mat
	};

	memcpy(scores,temp,6*6*sizeof(short));
	alignment_length = (corr_length + read_length+1)  ;

	SetForScoreing();

//	read_length = Config.GetInt("qry_max_len"); //Read Size
//	corr_length = Config.GetInt("corridor") + 1;//Corridor Size
//	ref_length = read_length + corr_length; //Reference Size
//
//	cigar = bool(((gpu_id >> 8) & 0xFF) == 1);
//
//	batch_size = 1;
//
//	mat = Config.GetFloat("score_match");
//	mis = Config.GetFloat("score_mismatch");
//	gap_read = Config.GetFloat("score_gap_read");
//	gap_ref = Config.GetFloat("score_gap_ref");
//
//	type = true;
//	short temp[6][6] = { mat, mis, mis, mis, 0, mis, mis, mat, mis, mis, 0, mis, mis, mis, mat, mis, 0, mis, mis, mis, mis, mat, 0, mis, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, mat };
//
//	memcpy(scores, temp, 6 * 6 * sizeof(short));
//	alignment_length = (corr_length + read_length + 1);
//
//	SetForScoreing();
}

SWCPUCor::~SWCPUCor() {
	FreeMem();
}

//SETTER:


void SWCPUCor::SetForScoreing() {
	if (!type) {
		FreeMem();
	}
	mem_matrix = corr_length;
	type = true;

	Allocate();
}

void SWCPUCor::SetForBacktracking() {

	if (type) {
		FreeMem();
	}
	type = false;

	mem_matrix = (Config.GetInt("corridor") + 2) * (Config.GetInt("qry_max_len") + 1);
	result = new short[result_number];
	mat_pointer = new char[mem_matrix];

	Allocate();
}

//====================== Backtracking ============================


int SWCPUCor::Calc_Alignment_SW_cigar(int const batchSize, char const * const * const refSeqList, char const * const * const qrySeqList, Align * results) {

	memset(mat_pointer, CIGAR_STOP, mem_matrix * sizeof(char));

	SW_Score(refSeqList[0], qrySeqList[0], result);

	short * alignments = new short[alignment_length * 2];
	memset(alignments, '\0', alignment_length * 2 * sizeof(short));

	int si = (Config.GetInt("corridor") + 2) * (Config.GetInt("qry_max_len") + 1);

	Backtracking_CIGAR(refSeqList[0], qrySeqList[0], result, alignments);

	computeCigarMD(results[0], result[3], alignments, refSeqList[0] + result[0]);
	results[0].PositionOffset = result[0];
	delete[] alignments;
	return batchSize;
}

int SWCPUCor::Calc_Alignment_SW(int const batchSize, char const * const * const refSeqList, char const * const * const qrySeqList, Align * results) {

	memset(mat_pointer, CIGAR_STOP, mem_matrix * sizeof(char));

	SW_Score(refSeqList[0], qrySeqList[0], result);

	char * alignments = new char[alignment_length * 2];
	memset(alignments, '\0', alignment_length * 2 * sizeof(char));
	Backtracking(refSeqList[0], qrySeqList[0], result, alignments);

	float total = 0.0f;
	float match = 0.0f;
	char * read = results[0].pQry;
	char * ref = results[0].pRef;

	int index = 0;
	for (int t = result[3] + 1; t < alignment_length; ++t) {
		ref[index] = alignments[t];
		read[index] = alignments[t + alignment_length];

		if (read[index] != ' ' && read[index] != '-') {
			total++;
		}
		if (read[index] != ' ' && read[index] == ref[index]) {
			match++;
		}
		index += 1;
	}

	results[0].PositionOffset = result[0];
	results[0].QStart = result[1];
	results[0].QEnd = result[2];
	results[0].Identity = match / total;

	delete[] alignments;
	return batchSize;
}

int SWCPUCor::Calc_Alignment_NW(int const batchSize, char const * const * const refSeqList, char const * const * const qrySeqList, Align * results) {
	cout<<"NW align"<<endl;
	memset(mat_pointer, CIGAR_STOP, mem_matrix * sizeof(char));

	NW_Score(refSeqList[0], qrySeqList[0], result);

	char * alignments = new char[alignment_length*2];
	NWBacktracking(refSeqList[0], qrySeqList[0], result,alignments);

	float total = 0.0f;
	float match = 0.0f;
	char * read = results[0].pQry;
	char * ref = results[0].pRef;

	int index = 0;
	for (int t = result[3] + 1; t < alignment_length; ++t) {
		ref[index] = alignments[t];
		read[index] = alignments[t + alignment_length];

		if (read[index] != ' ' && read[index] != '-') {
			total++;
		}
		if (read[index] != ' ' && read[index] == ref[index]) {
			match++;
		}
		index += 1;
	}

	results[0].PositionOffset = result[0];
	results[0].QStart = result[1];
	results[0].QEnd = result[2];
	results[0].Identity = match / total;

	delete [] alignments;
	return batchSize;
}

int printCigarElement(char const op, short const length, char * cigar) {
	int offset = 0;
	offset = sprintf(cigar, "%d%c", length, op);
	return offset;
}

void SWCPUCor::computeCigarMD(Align & result, int const gpuCigarOffset, short const * const gpuCigar, char const * const refSeq) {
	int alignment_length = corr_length + read_length+1;


	int cigar_offset = 0;
	int md_offset = 0;

	if ((gpuCigar[gpuCigarOffset] >> 4) > 0) {
		cigar_offset += printCigarElement('S', gpuCigar[gpuCigarOffset] >> 4, result.pRef + cigar_offset);
		result.QStart = gpuCigar[gpuCigarOffset] >> 4;
	}

	int cigar_m_length = 0;
	int md_eq_length = 0;
	int ref_index = 0;
	for (int j = gpuCigarOffset + 1; j < (alignment_length - 1); ++j) {
		int op = gpuCigar[j] & 15;
		int length = gpuCigar[j] >> 4;

		//debugCigar(op, length);

		switch (op) {
			case CIGAR_X:
				cigar_m_length += length;

				//Produces: 	[0-9]+(([A-Z]+|\^[A-Z]+)[0-9]+)*
				//instead of: 	[0-9]+(([A-Z]|\^[A-Z]+)[0-9]+)*
				md_offset += sprintf(result.pQry + md_offset, "%d", md_eq_length);
				for (int k = 0; k < length; ++k) {
					md_offset += sprintf(result.pQry + md_offset, "%c", refSeq[ref_index++]);
				}
				md_eq_length = 0;

				break;
			case CIGAR_EQ:
				cigar_m_length += length;
				md_eq_length += length;
				ref_index += length;
				break;
			case CIGAR_D:
				if (cigar_m_length > 0) {
					cigar_offset += printCigarElement('M', cigar_m_length, result.pRef + cigar_offset);
					cigar_m_length = 0;
				}
				cigar_offset += printCigarElement('D', length, result.pRef + cigar_offset);

				md_offset += sprintf(result.pQry + md_offset, "%d", md_eq_length);
				md_eq_length = 0;
				result.pQry[md_offset++] = '^';
				for (int k = 0; k < length; ++k) {
					result.pQry[md_offset++] = refSeq[ref_index++];
				}

				break;
			case CIGAR_I:
				if (cigar_m_length > 0) {
					cigar_offset += printCigarElement('M', cigar_m_length, result.pRef + cigar_offset);
					cigar_m_length = 0;
				}
				cigar_offset += printCigarElement('I', length, result.pRef + cigar_offset);

				break;
			default:
				Log.Error("Invalid cigar string: %d", op);
				std::cout << "Offset: " << gpuCigarOffset << std::endl;
				for(int x = 0; x < alignment_length * 2; ++x) {
					std::cout << gpuCigar[x] << " ";
				}
				std::cout << std::endl;
				exit(1);
		}
	}
	md_offset += sprintf(result.pQry + md_offset, "%d", md_eq_length);
	if (cigar_m_length > 0) {
		cigar_offset += printCigarElement('M', cigar_m_length, result.pRef + cigar_offset);
		cigar_m_length = 0;
	}

	if ((gpuCigar[alignment_length - 1] >> 4) > 0) {
		cigar_offset += printCigarElement('S', gpuCigar[alignment_length - 1] >> 4, result.pRef + cigar_offset);
		result.QEnd = gpuCigar[alignment_length - 1] >> 4;
	}

	result.pRef[cigar_offset] = '\0';
	result.pQry[md_offset] = '\0';
}



void SWCPUCor::NWBacktracking( char const * scaff, char const * read,  short *& result,char *& alignments) {

	char * matrix_pointer= mat_pointer;
	//Index index;
	short best_read_index = result[0];
	short best_ref_index = result[1];
	if (best_read_index > 0) {
		matrix_pointer += (((corr_length + 1) * (best_read_index + 1)));

		short abs_ref_index = best_ref_index + best_read_index;
		short alignment_index =alignment_length - 2;

		alignments[alignment_length-1] = '\0';
		alignments[alignment_length-1 + alignment_length] = '\0';

		int pointer = CIGAR_STOP;
		while ((pointer =  matrix_pointer[best_ref_index + 1]) != CIGAR_STOP ) {

			if (pointer == CIGAR_X || pointer == CIGAR_EQ) {
				alignments[alignment_index] = scaff[abs_ref_index--];
				alignments[alignment_index + alignment_length] = read[best_read_index];
				matrix_pointer -= ((corr_length + 1));
				best_read_index -= 1;
			} else if (pointer == CIGAR_I) {
				alignments[alignment_index] = '-';
				alignments[alignment_index + alignment_length] = read[best_read_index];
				matrix_pointer -= ((corr_length + 1));
				best_read_index -= 1;
				best_ref_index += 1;
			} else{
				alignments[alignment_index] = scaff[abs_ref_index--];
				alignments[alignment_index + alignment_length] = '-';
				best_ref_index -= 1;
			}
			alignment_index -= 1;
		}
		result[0] = abs_ref_index + 1;
		result[1] = best_read_index + 1;
		result[3] = alignment_index;
	}
}


void SWCPUCor::Backtracking_CIGAR(char const * scaff, char const * read, short *& result, short *& alignments) {

	char * matrix_pointer= mat_pointer;

	short best_read_index = result[param_best_read_index];
	short best_ref_index = result[param_best_ref_index];

	if (best_read_index > 0) {
		matrix_pointer += (((corr_length + 1) * (best_read_index + 1)));

		short abs_ref_index = best_ref_index + best_read_index;
		short alignment_index = alignment_length  - 1;

		int pointer = CIGAR_STOP;
		int cigar_element = CIGAR_S;
		int cigar_length = result[qend];
		while ((pointer = matrix_pointer[(best_ref_index + 1)]) != CIGAR_STOP) {

			if (pointer == CIGAR_X || pointer == CIGAR_EQ) {
				matrix_pointer -= ((corr_length + 1));
				best_read_index -= 1;
				abs_ref_index -= 1;
			} else if (pointer == CIGAR_I) {
				matrix_pointer -= ((corr_length + 1));
				best_read_index -= 1;
				best_ref_index += 1;
			} else {
				best_ref_index -= 1;
				abs_ref_index -= 1;
			}

			if (pointer == cigar_element) {
				cigar_length += 1;
			} else {
				alignments[alignment_index--] = (cigar_length << 4 | cigar_element);
				cigar_element = pointer;
				cigar_length = 1;
			}
		}
		alignments[alignment_index--] = (cigar_length << 4 | cigar_element);
		alignments[alignment_index] = ((best_read_index + 1) << 4 | CIGAR_S);
		result[ref_position] = abs_ref_index + 1;
		result[qstart] = best_read_index + 1;
		//qend was set by "forward" kernel
		result[alignment_offset] = alignment_index;
	}
}


int SWCPUCor::Calc_Alignment_NW_cigar(int const batchSize, char const * const * const refSeqList, char const * const * const qrySeqList, Align * results) {
	memset(mat_pointer, CIGAR_STOP, mem_matrix * sizeof(char));

	NW_Score(refSeqList[0],qrySeqList[0],result);

	short * alignments = new short[alignment_length*2];
	memset(alignments,'\0',alignment_length*2*sizeof(short));

	int si=(Config.GetInt("corridor") + 2) * (Config.GetInt("qry_max_len") + 1);

	Backtracking_CIGAR(refSeqList[0], qrySeqList[0], result,alignments);

	computeCigarMD(results[0], result[3], alignments, refSeqList[0] + result[0]);
	results[0].PositionOffset = result[0];
	delete [] alignments;
	return batchSize;
}

int SWCPUCor::GetScoreBatchSize() {
	return sse->GetScoreBatchSize();
}
int SWCPUCor::GetAlignBatchSize() {
	return batch_size;
}

int SWCPUCor::BatchAlign(int const mode, int const batchSize, char const * const * const refSeqList, char const * const * const qrySeqList, Align * const results, void * extData) {
	if (batchSize <= 0) {
		Log.Warning("Align for batchSize <= 0");
		return 0;
	}

	if (type) {
		SetForBacktracking();
	}

	switch (mode & 0xFF) {
	case 0:
		if (cigar) {
			return Calc_Alignment_SW_cigar(batchSize, refSeqList, qrySeqList, results);
		} else {
			return Calc_Alignment_SW(batchSize, refSeqList, qrySeqList, results);
		}

	case 1:
		if (cigar) {
			return Calc_Alignment_NW_cigar(batchSize, refSeqList, qrySeqList, results);
		} else {
			return Calc_Alignment_NW(batchSize, refSeqList, qrySeqList, results);
		}
	default:
		return 0;
	}

	Log.Error("Unsupported alignment mode %i", mode);
	return 0;
}

int SWCPUCor::BatchScore(int const mode, int const batchSize, char const * const * const refSeqList, char const * const * const qrySeqList, float * const results, void * extData) {
	if (batchSize <= 0) {
		Log.Warning("Score for batchSize <= 0");
		return 0;
	}

	if (!type) {
		SetForScoreing();
	}
	int b1 = batchSize % 8;


	for (int i = 0; i < b1; ++i) {
		switch (mode & 0xFF) {
				case 0:
					Log._Message(1,"first score","SW");
					CalcScores_SW(1, &refSeqList[i], &qrySeqList[i], &results[i]);
					break;
				case 1:
					Log._Message(1,"first score","NW");
					CalcScores_NW(1, &refSeqList[i], &qrySeqList[i], &results[i]);
					break;
				default:
					Log.Error("Unsupported alignment mode %i", mode & 0xFF);
					return 0;
				}
	}

	if(batchSize>b1){
		return b1+ sse->BatchScore(mode, batchSize-b1, &refSeqList[b1], &qrySeqList[b1], &results[b1], 0);
	}else{
		return b1;
	}

}

